"""
------------------------------------------
cm_lib: utilities for handling Cm-messages
------------------------------------------
"""
# B. Swinkels, 2014

from future import standard_library
standard_library.install_aliases()
from builtins import object, str
import PyCm as cm
import logging
import numpy as np
import subprocess as sp
import ctypes
import queue
import time
import os
from virgotools.pvt_utils import ascii, native_str


__all__ = ['cm_handler', 'CmError', 'Char', 'cm_start', 'cm_start_unique',
           'cm_send', 'cm_send_receive']


logger = logging.getLogger(__name__)

cm_handler = cm.CmMessageHandler
"""decorator for giving the right ctypes signature for call-back functions of
cm-handlers"""


class CmError(Exception):
    pass


class Char(object):
    """container for passing a char to :func:`cm_send`

    Examples
    --------
    >>> cm_send('some_server', 'some_function', 1, 2.34, 'some_string', Char('x'))

    """
    def __init__(self, char):
        if not (isinstance(char, (str, bytes)) and len(char) == 1):
            raise ValueError('Char needs a str of length 1')
        self.char = ascii(char)

    def __repr__(self):
        return 'Char({!r})'.format(self.char)

    def __eq__(self, other):
        return self.char == other.char


_my_cm_name = None  # global state, also used by FdIO


def cm_start(cm_name):
    """
    Register with given cm_name. Not needed for processes that do use FdIO. Does nothing
    if you already have a cm_name.

    Parameters
    ----------
    cm_name : str
        requested name

    Raises
    ------
    CmError
        If process can not register with given cm-name.
    """
    cm_name = ascii(cm_name)
    global _my_cm_name
    if _my_cm_name is None:
        logger.info("Starting with cm-name %s", cm_name)
        ok = cm.CmMessageOpenServer(cm_name)
        if not ok:
            msg = "Could not register with cm-name %s" % cm_name
            raise CmError(msg)
        _my_cm_name = cm_name
    else:
        logger.info("Called cm_start, but I already have a cm-name %s", _my_cm_name)


def cm_start_unique(prefix):
    """
    Register with a unqiue cm-name. The name is formed by the prefix, username,
    host and time to guarantee a unique name. If you are already register with
    a cm-name (e.g when reloading a script from within ipython) this fuction
    does nothing.

    Parameters
    ----------
    prefix : str
        prefix of the cm_name
    """
    if _my_cm_name is None:
        cm_name = '{}_{}_{}_{}'.format(
            prefix,
            os.getenv('USER', 'UNKNOWN_USER'),
            os.getenv('HOSTNAME', 'UNKNOWN_HOST').split('.')[0],
            time.strftime("%Y%m%d_%H%M%M", time.gmtime()))
        cm_start(cm_name)
    else:
        logger.info("Called cm_start_unique, but I already have a cm-name %s",
                    _my_cm_name)


def cm_send(dest, cm_type, *args):
    r"""
    Send Cm message.

    Send a Cm message to different process of given type and arguments. Before
    using this function, the current process must be registered with its own
    cm-name (either via cm_start or FdIO). For the allowable values for
    cm_type and the type of the arguments, consult the manual of the destination
    process.

    Parameters
    ----------
    dest : str
        cm-name of destination
    cm_type : str
        type of cm-command, see the manual of the destination process
    \*args
        Arguments of cm-message. Note that the type of the arguments is taken
        from the type of the Python argument. For text arguments, use Python
        str. For Int use any Python or Numpy integer. For Double use Python
        floats or numpy.float64. For Float arguments, use numpy.float32. To
        send a char pass a :class:`.Char` object.

    Raises
    ------
    CmError
        If message can not be sent, e.g. when the current process does not have
        a cm-name, or if the destination process is not reachable.

    Examples
    --------
    >>> cm_start('MyCmName')
    >>> cm_send('PrISYS', 'PrSwTranSet', 'LOOP_ENBL', -1, float(ramp_time))
    """
    dest = ascii(dest)
    cm_type = ascii(cm_type)

    arg_str = [repr(dest), repr(cm_type)]
    message = cm.CmMessageNew()
    assert message, 'could not create CmMessage'
    try:
        cm.CmMessageSetType(message, cm_type)
        for arg in args:
            if isinstance(arg, (str, bytes)):
                arg = ascii(arg)
                cm.CmMessagePutText(message, arg)
                arg_str.append(repr(arg))
            elif isinstance(arg, Char):
                cm.CmMessagePutChar(message, arg.char)
                arg_str.append(repr(arg))
            elif isinstance(arg, (int, np.integer)):
                cm.CmMessagePutInt(message, arg)
                arg_str.append(repr(arg))
            elif isinstance(arg, np.float32):
                cm.CmMessagePutFloat(message, arg)
                arg_str.append('np.float32({!r})'.format(arg))
            elif isinstance(arg, (float, np.float64)):
                cm.CmMessagePutDouble(message, arg)
                arg_str.append(repr(arg))
            else:
                msg = 'cm_send: cannot handle arguments of type {}'.format(
                    type(arg))
                raise TypeError(msg)
        logger.info('cm_send(%s)', ', '.join(arg_str))
        res = cm.CmMessageSend(message, dest)
        if not res:
            if _my_cm_name is None:
                raise CmError("I do not have a cm-name, use cm_start first")
            else:
                msg = "Could not send message, check if %s exists." % dest
                raise CmError(msg)
    finally:
        cm.CmMessageDelete(message)


CM_COMMAND = '$CMROOT/$UNAME/cm.exe'


def cm_send_cli(dest, cm_type, *args):
    """Obsolete cm_send that uses the command line tool, use cm_send instead"""
    command = [CM_COMMAND, 'send -to', dest, '-type', cm_type]
    for arg in args:
        if isinstance(arg, (str, bytes)):
            command.append('-t "%s"' % arg)
        elif isinstance(arg, Char):
            command.append('-c %s' % arg.char)
        elif isinstance(arg, (int, np.integer)):
            command.append('-i %i' % arg)
        elif isinstance(arg, np.float32):
            command.append('-f %r' % arg)  # use repr to avoid losing precision
        elif isinstance(arg, (float, np.float64)):
            command.append('-d %r' % arg)  # use repr to avoid losing precision
        else:
            raise TypeError(
                'cm_send: expected string, int or float arguments, got %s' %
                type(arg))
    command = ' '.join(command)
    logger.info('Sending cm-command: %s', command)
    try:
        sp.check_call(command, shell=True)
    except sp.CalledProcessError:
        logger.error('Error while sensing cm-command: %s', command)


_CM_TYPE = {  # CmMessageGetItemType -> appropriate getter and converter
    cm.CmMessageItemTail: (None, None),
    cm.CmMessageItemChar: (cm.CmMessageGetChar, bytes),
    cm.CmMessageItemShort: (cm.CmMessageGetShort, int),
    cm.CmMessageItemInt: (cm.CmMessageGetInt, int),
    cm.CmMessageItemLong: (cm.CmMessageGetLong, int),
    cm.CmMessageItemFloat: (cm.CmMessageGetFloat, float),
    cm.CmMessageItemDouble: (cm.CmMessageGetDouble, float),
    cm.CmMessageItemText: (cm.CmMessageGetText, native_str),
    cm.CmMessageItemBytes: (cm.CmMessageGetBytes, NotImplemented),
    cm.CmMessageItemArray: (cm.CmMessageGetArray, NotImplemented),
}


def parse_cm_message(message):
    """Extract arguments from cm-message and return as list of Python objects"""
    items = []
    while True:
        getter, converter = _CM_TYPE[cm.CmMessageGetItemType(message)]
        if getter is None:
            break
        item = getter(message)  # weird error with "cm send ... -i 1.23"
        item = converter(item)
        items.append(item)
    return items


class CmReceiver(object):
    """class for receiving a single Cm reply. Can be used as context manager for
    installing and uninstalling handler, see cm_send_receive for usage"""
    def __init__(self, receive_type):
        self.receive_type = receive_type
        self.message = None

        @cm_handler
        def handler(message, sender, _server):
            msg_type = native_str(cm.CmMessageGetType(message))
            sender = native_str(sender)
            logger.info('Received %s message from %s',
                        msg_type, sender)

            if not self.receive_type == msg_type:
                raise CmError('Got message of type %s, but expected %s' %
                              (msg_type, self.receive_type))

            self.message = parse_cm_message(message)
            return cm.CmMessageBreak

        self.handler = handler

    def __enter__(self):
        logger.debug('Installing handler for %s', self.receive_type)
        cm.CmMessageInstallHandler(self.handler, ascii(self.receive_type))
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        logger.debug('Uninstalling handler for %s', self.receive_type)
        cm.CmMessageUninstallHandler(ascii(self.receive_type))

    def await_reply(self, timeout):
        self.message = None  # in case the object is reused
        twait = timeout
        tstart = time.time()

        while True:
            result = cm.CmMessageWaitWithTimeout(twait)
            twaited = time.time() - tstart

            if result == cm.CmConnectNoHandler:
                raise CmError('got CmConnectNoHandler')
            if result == cm.CmConnectTimeoutDetection:
                msg = 'Did not receive message after %g seconds' % timeout
                raise CmError(msg)
            if result == cm.CmConnectErrorCondition:
                raise CmError('got CmConnectErrorCondition')
            assert result == cm.CmConnectBreakDetection

            if self.message is not None:  # message received
                break

            twait = timeout - twaited
            logger.info('Waited %g s but got no message, waiting for %g s more',
                        twaited, twait)

        logger.debug('Got message in %g s', twaited)
        return self.message


def cm_send_receive(receive_type, timeout, dest, send_type, *args):
    r"""
    Send cm-message and receive reply.

    Parameters
    ----------
    receive_type : str
        Expected type of received message.
    timeout : float
        How long to wait for the reply.
    dest, send_type, \*args
        Destination, type and arguments for sent message, see :func:`cm_send`.

    Returns
    -------
    list
        List of various objects, based on type of Cm message.

    Raises
    ------
    CmError
        If something goes wrong, e.g. the message cannot be sent, or if no reply
        is received after given timeout.

    Example
    -------
    >>> cm_start('MyCmName')
    >>> names = cm_send_receive('NSNames', 1, 'CmNameServer', 'NSGetNames')
    """
    with CmReceiver(receive_type) as receiver:
        cm_send(dest, send_type, *args)
        message = receiver.await_reply(timeout)
    return message
